#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Anna.h"
#include "Bruno.h"

typedef long long ll;
using namespace std;

namespace {
	FILE *in, *out;

	enum {
		ERR_TLENGTH = 1,
		ERR_TCHAR = 2,
		ERR_TCOINSIDE = 3,
		ERR_LRANGE = 4,
		ERR_BRUNOCORRECT = 5
	};

	int MAXL = 130;

} // namespace

void wrong(int num) {
	fprintf(out, "%d\n", num);
	fflush(out);
	exit(0);  // do not exit(1)! (it might cause RE where WA is appropriate)
}

void check_manager_error(){
	char flag[20];
	if (fscanf(in, "%10s%*[^\n]", flag) != 1) {
		fprintf(stderr, "user program 0 failed to read data from manager");
		wrong(-1);
	}
	if (strcmp("WA", flag) == 0) { // (flag == "WA")
		exit(0);
	} else if (strcmp("Success", flag) != 0) { // (flag != "Success")
		fprintf(stderr, "wrong manager message to user program 0\n");
		wrong(-1);
	}
}


int main(int argc, char** argv) {
	if (argc < 4) {
		fprintf(stderr, "Usage: %s <in> <out> <shard>\n", argv[0]);
		return 1;
	}

	// open PIPE
	signal(SIGPIPE, SIG_IGN);
	out = fopen(argv[2], "w");
	in  = fopen(argv[1], "r");

	// 2.prepare
	char prepare[20];
	if (fscanf(in, "%10s%*[^\n]", prepare) != 1) {
		fprintf(stderr, "program failed to read data from manager\n");
		exit(1);
	}
	// 3.prepare
	fprintf(out, "Start\n");
	fflush( out);

	int shard = atoi(argv[3]);
	if (shard == 0) {
		// anna
		
		// 6.read response from manager
		check_manager_error();
		
		// 7.write error code OK
		fprintf(out, "%d\n", 0);
		fflush( out);


		// 10.read data (N,S) from manager
		int N;
		if (fscanf(in, "%d", &N) != 1) {
			fprintf(stderr, "user program 0 failed to read data N from manager");
			wrong(-1);
		}
		char ch;
		if (fscanf(in,"%c",&ch)!=1){
			fprintf(stderr, "user program 0 failed to read LF from manager");
			wrong(-1);
		}
		char buf[500001];
		if (fscanf(in, "%500000s%*[^\n]", buf) != 1) {
			fprintf(stderr, "user program 0 failed to read data S from manager");
			wrong(-1);
		}

		string S=buf;

		// call anna
		auto [t,l]= anna(N,S);

		// validate (Wrong 1,2,3,4)
		if((int)t.size()!=N) wrong( (int)ERR_TLENGTH);

		for(auto c:t){
			if(c!='R'&&c!='G'&&c!='B'){
				wrong((int)ERR_TCHAR);
			}
		}

		for(int i=0;i<N;i++){
			if(S[i]==t[i]){
				wrong((int)ERR_TCOINSIDE);
			}
		}

		if (!(1 <= l && l <= min(N,MAXL))) {
			wrong((int)ERR_LRANGE);
		}

		// 11.write error code OK
		fprintf(out, "%d\n", 0);
		fflush( out);

		// 13.write data (t, l) to manager
		for(int i=0;i<N;i++){
			fprintf(out, "%c", t[i]);
		}
		fprintf(out, "\n");
		fflush(out);
		fprintf(out, "%d\n", l);
		fflush(out);

	} else if (shard == 1) {
		// bruno

		// 16.read response from manager
		check_manager_error();

		// 18.read data (Q,N,l) from manager
		int Q,N,l;
		if (fscanf(in, "%d", &Q) != 1) {
			fprintf(stderr, "user program 1 failed to read Q from manager");
			wrong(-1);
		}
		if (fscanf(in, "%d", &N) != 1) {
			fprintf(stderr, "user program 1 failed to read N from manager");
			wrong(-1);
		}
		if (fscanf(in, "%d", &l) != 1) {
			fprintf(stderr, "user program 1 failed to read L from manager");
			wrong(-1);
		}
		// 19.write error code OK
		fprintf(out, "%d\n", 0);
		fflush( out);

		//call init()
		init(N,l);

		for (int i = 0; i < Q; i++) {
			// q2.read data (u) from manager
			char ch;
			if(fscanf(in,"%c",&ch)!=1){
				fprintf(stderr, "user program 1 failed to read LF from manager");
				wrong(-1);
			}
			char buf[131];
			if (fscanf(in, "%130s%*[^\n]", buf) != 1) {
				fprintf(stderr, "user program 1 failed to read u_%d from manager",i);
				wrong(-1);
			}
			string u(buf,l);

			if((int)u.size()!=l){
				fprintf(stderr, "invalid u length");
				wrong(-1);
			}
			for(int i=0;i<l;i++){
				if(!(u[i]=='R'||u[i]=='G'||u[i]=='B')){
					fprintf(stderr,"invalid u characters");
					wrong(-1);
				}
			}

			// call bruno
			int guess = bruno(u);

			// q3.write error code OK
			fprintf(out, "%d\n", 0);
			fflush( out);
			
			// q5.write data (guess) to manager
			fprintf(out, "%d\n", guess);
			fflush( out);

			// q8.read response from manager
			check_manager_error();
		}

	}

	return 0;
}
